@@ -28,10 +28,8 @@
 #include "qdjview.h"
 #include "qdjvuwidget.h"
 #include "djview.h"
-#ifdef Q_WS_X11
-# ifndef NPDJVU
-#  include "qdjviewplugin.h"
-# endif
+#ifndef NPDJVU
+# include "qdjviewplugin.h"
 #endif
 
 #include <QtGlobal>
@@ -49,16 +47,15 @@
 #include <QStringList>
 #include <QTranslator>
 
-
 #ifdef QT_NO_DEBUG
 static bool verbose = false;
 #else
 static bool verbose = true;
 #endif
-
+static bool appReady = false;
 
 static void 
-qtMessageHandler(QtMsgType type, const char *msg)
+qt4MessageHandler(QtMsgType type, const char *msg)
 {
   switch (type) 
     {
@@ -69,16 +66,25 @@ qtMessageHandler(QtMsgType type, const char *msg)
       fprintf(stderr,"djview critical error: %s\n", msg);
       break;
     case QtWarningMsg:
-      if (verbose)
+      if (verbose || !appReady)
         fprintf(stderr,"djview: %s\n", msg);
       break;
     default:
-      if (verbose)
+      if (verbose || !appReady)
         fprintf(stderr,"%s\n", msg);
       break;
     }
 }
 
+#if QT_VERSION >= 0x50000
+static void 
+qt5MessageHandler(QtMsgType type, const QMessageLogContext&, const QString &msg)
+{
+  QByteArray bmsg = msg.toLocal8Bit();
+  qt4MessageHandler(type, bmsg.data());
+}
+#endif
+
 
 static void
 message(QString string, bool prefix=true)
@@ -96,22 +102,16 @@ message(QStringList sl)
     message(s);
 }
 
-
-static void
-addDirectory(QStringList &dirs, QString path)
-{
-  QString dirname = QDir::cleanPath(path);
-  if (! dirs.contains(dirname))
-    dirs << dirname;
-}
-
-
 QDjViewApplication::QDjViewApplication(int &argc, char **argv)
   : QApplication(argc, argv), 
     context(argv[0])
 {
   // Message handler
-  qInstallMsgHandler(qtMessageHandler);
+#if QT_VERSION >= 0x50000
+  qInstallMessageHandler(qt5MessageHandler);
+#else
+  qInstallMsgHandler(qt4MessageHandler);
+#endif
   
   // Locale should not mess with printf
   // We do this again because old libdjvulibre
@@ -124,87 +124,31 @@ QDjViewApplication::QDjViewApplication(int &argc, char **argv)
 #endif
   
   // Mac/Cocoa bug workaround
-#if defined(Q_WS_MAC) && defined(QT_MAC_USE_COCOA) && QT_VERSION<0x40503
+#if defined(Q_OS_DARWIN) && defined(QT_MAC_USE_COCOA) && QT_VERSION<0x40503
   extern void qt_mac_set_native_menubar(bool);
   qt_mac_set_native_menubar(false);
 #endif
-  
-  // Translators
-  QTranslator *qtTrans = new QTranslator(this);
-  QTranslator *djviewTrans = new QTranslator(this);
-  // - determine preferred languages
-  QStringList langs; 
-  QString langOverride = QSettings().value("language").toString();
-  if (! langOverride.isEmpty())
-    langs += langOverride;
-  QString varLanguage = ::getenv("LANGUAGE");
-  if (varLanguage.size())
-    langs += varLanguage.toLower().split(":", QString::SkipEmptyParts);
-#ifdef LC_MESSAGES
-  QString varLcMessages = ::setlocale(LC_MESSAGES, 0);
-  if (varLcMessages.size())
-    langs += varLcMessages.toLower();
-#else
-# ifdef LC_ALL
-  QString varLcMessages = ::setlocale(LC_ALL, 0);
-  if (varLcMessages.size())
-    langs += varLcMessages.toLower();
-# endif
-#endif
-#ifdef Q_WS_MAC
-  langs += QSettings(".", "globalPreferences")
-    .value("AppleLanguages").toStringList();
-#endif
-  QString qtLocale =  QLocale::system().name();
-  if (qtLocale.size())
-    langs += qtLocale.toLower();
 
-  // - determine potential directories
-  QStringList dirs;
-  QDir dir = applicationDirPath();
-  QString dirPath = dir.canonicalPath();
-  addDirectory(dirs, dirPath);
-#ifdef DIR_DATADIR
-  QString datadir = DIR_DATADIR;
-  addDirectory(dirs, datadir + "/djvu/djview4");
-  addDirectory(dirs, datadir + "/djview4");
-#endif
-#ifdef Q_WS_MAC
-  addDirectory(dirs, dirPath + "/Resources/$LANG.lproj");
-  addDirectory(dirs, dirPath + "/../Resources/$LANG.lproj");
-  addDirectory(dirs, dirPath + "/../../Resources/$LANG.lproj");
-#endif
-  addDirectory(dirs, dirPath + "/share/djvu/djview4");
-  addDirectory(dirs, dirPath + "/share/djview4");
-  addDirectory(dirs, dirPath + "/../share/djvu/djview4");
-  addDirectory(dirs, dirPath + "/../share/djview4");
-  addDirectory(dirs, dirPath + "/../../share/djvu/djview4");
-  addDirectory(dirs, dirPath + "/../../share/djview4");
-  addDirectory(dirs, "/usr/share/djvu/djview4");
-  addDirectory(dirs, "/usr/share/djview4");
-  addDirectory(dirs, QLibraryInfo::location(QLibraryInfo::TranslationsPath));
-  // - load translators
-  bool qtTransValid = false;
-  bool djviewTransValid = false;
-  foreach (QString lang, langs)
+  // Wire session management signals
+  connect(this, SIGNAL(saveStateRequest(QSessionManager&)),
+          this, SLOT(saveSessionState(QSessionManager&)) );
+  
+  // Install translators
+  QStringList langs = getTranslationLangs();
+  QTranslator *djviewTrans = new QTranslator(this);
+  QTranslator *qtTrans = new QTranslator(this);
+  if (loadTranslators(langs, djviewTrans, qtTrans))
     {
-      foreach (QString dir, dirs)
-        {
-          dir = dir.replace(QRegExp("\\$LANG(?!\\w)"),lang);
-          QDir qdir(dir);
-          if (! qtTransValid && qdir.exists())
-            qtTransValid = qtTrans->load("qt_" + lang, dir, "_.-");
-          if (! djviewTransValid && qdir.exists())
-            djviewTransValid= djviewTrans->load("djview_" + lang, dir, "_.-");
-        }
-      if (lang == "en" || lang.startsWith("en_") || lang == "c")
-        break;
+      installTranslator(qtTrans);
+      installTranslator(djviewTrans);
     }
-  // - install tranlators
-  if (qtTransValid)
-    installTranslator(qtTrans);
-  if (djviewTransValid)
-    installTranslator(djviewTrans);
+  else
+    {
+      delete qtTrans;
+      delete djviewTrans;
+    }
+  // Application is ready
+  appReady = true;
 }
 
 
@@ -220,6 +164,131 @@ QDjViewApplication::newWindow()
 }
 
 
+static void
+addDirectory(QStringList &dirs, QString path)
+{
+  QString dirname = QDir::cleanPath(path);
+  if (! dirs.contains(dirname))
+    dirs << dirname;
+}
+
+
+QStringList 
+QDjViewApplication::getTranslationDirs()
+{
+  if (translationDirs.isEmpty())
+    {
+      QStringList dirs;
+      QDir dir = applicationDirPath();
+      QString dirPath = dir.canonicalPath();
+      addDirectory(dirs, dirPath);
+#ifdef DIR_DATADIR
+      QString datadir = DIR_DATADIR;
+      addDirectory(dirs, datadir + "/djvu/djview4");
+      addDirectory(dirs, datadir + "/djview4");
+#endif
+#ifdef Q_OS_DARWIN
+      addDirectory(dirs, dirPath + "/Resources/$LANG.lproj");
+      addDirectory(dirs, dirPath + "/../Resources/$LANG.lproj");
+      addDirectory(dirs, dirPath + "/../../Resources/$LANG.lproj");
+#endif
+      addDirectory(dirs, dirPath + "/share/djvu/djview4");
+      addDirectory(dirs, dirPath + "/share/djview4");
+      addDirectory(dirs, dirPath + "/../share/djvu/djview4");
+      addDirectory(dirs, dirPath + "/../share/djview4");
+      addDirectory(dirs, dirPath + "/../../share/djvu/djview4");
+      addDirectory(dirs, dirPath + "/../../share/djview4");
+      addDirectory(dirs, "/usr/share/djvu/djview4");
+      addDirectory(dirs, "/usr/share/djview4");
+      addDirectory(dirs, QLibraryInfo::location(QLibraryInfo::TranslationsPath));
+      translationDirs = dirs;
+    }
+  return translationDirs;
+}
+
+
+static void
+addLang(QStringList &langs, QString s)
+{
+  if (s.size() > 0)
+    {
+      s = s.replace(QChar('-'), QChar('_'));
+      if (! langs.contains(s))
+        langs << s;
+#ifdef Q_OS_DARWIN
+      if (s == "zh_Hans" && ! langs.contains("zh_CN"))
+        langs << "zh_CN";
+      if (s == "zh_Hant" && ! langs.contains("zh_TW"))
+        langs << "zh_TW";
+#endif
+    }
+}
+
+
+QStringList 
+QDjViewApplication::getTranslationLangs()
+{
+  if (translationLangs.isEmpty())
+    {
+      QStringList langs; 
+      addLang(langs, QSettings().value("language").toString());
+      QString varLanguage = QString::fromLocal8Bit(::getenv("LANGUAGE"));
+      foreach(QString lang, varLanguage.split(":", QString::SkipEmptyParts))
+        addLang(langs, lang);
+#ifdef LC_MESSAGES
+      addLang(langs, QString::fromLocal8Bit(::setlocale(LC_MESSAGES, 0)));
+#endif
+      addLang(langs, QString::fromLocal8Bit(::getenv("LANG")));
+#ifdef Q_OS_DARWIN
+      QSettings g(".", "globalPreferences");
+      foreach (QString lang, g.value("AppleLanguages").toStringList())
+        addLang(langs, lang);
+#endif
+      addLang(langs, QLocale::system().name());
+      translationLangs = langs;
+    }
+  return translationLangs;
+}
+
+
+static bool loadOneTranslator(QTranslator *trans, 
+                              QString name, QString lang, QStringList dirs)
+{
+  QString llang = lang.toLower();
+  foreach (QString dir, dirs)
+    {
+      dir = dir.replace(QRegExp("\\$LANG(?!\\w)"), lang);
+      QDir qdir(dir);
+      if (qdir.exists())
+        {
+          if (trans->load(name + "_" + lang, dir, "_.-"))
+            return true;
+          if (lang != llang && trans->load(name + "_" + llang, dir, "_.-"))
+            return true;
+        }
+    }
+  return false;
+}
+
+bool
+QDjViewApplication::loadTranslators(QStringList langs, 
+                                    QTranslator *dTrans, QTranslator *qTrans)
+{
+  QStringList dirs = getTranslationDirs();
+  foreach (QString lang, langs)
+    {
+      bool okay = true;
+      if (okay && dTrans && !loadOneTranslator(dTrans, "djview", lang, dirs))
+        okay = false;
+      if (okay && qTrans && !loadOneTranslator(qTrans, "qt", lang, dirs))
+        okay = false;
+      if (okay || lang.startsWith("en_") || lang == "en")
+        return okay;
+    }
+  return false;
+}
+
+
 bool 
 QDjViewApplication::event(QEvent *ev)
 {
@@ -245,16 +314,8 @@ QDjViewApplication::event(QEvent *ev)
   return QApplication::event(ev);
 }
 
-
-#ifdef Q_WS_X11
-
 void 
-QDjViewApplication::commitData(QSessionManager &)
-{
-}
-
-void 
-QDjViewApplication::saveState(QSessionManager &sm)
+QDjViewApplication::saveSessionState(QSessionManager &sm)
 {
   int n = 0;
   QSettings s;
@@ -291,8 +352,6 @@ QDjViewApplication::saveState(QSessionManager &sm)
     }
 }
 
-#endif
-
 
 #ifndef NPDJVU
 
@@ -329,6 +388,7 @@ usage()
   exit(10);
 }
 
+
 int 
 main(int argc, char *argv[])
 {
@@ -338,21 +398,35 @@ main(int argc, char *argv[])
   QApplication::setApplicationName(DJVIEW_APP);
   
   // Message handler
-  qInstallMsgHandler(qtMessageHandler);
+#if QT_VERSION >= 0x50000
+  qInstallMessageHandler(qt5MessageHandler);
+#else
+  qInstallMsgHandler(qt4MessageHandler);
+#endif
+
+  // Set verbose flag as early as possible
 #ifdef Q_OS_UNIX
   const char *s = ::getenv("DJVIEW_VERBOSE");
   if (s && strcmp(s,"0"))
     verbose = true;
 #endif
+  for (int i=1; i<argc; i++)
+    {
+      char *s = argv[i];
+      while (s && *s=='-') 
+        s++;
+      if (! strcmp(s, "verbose"))
+        verbose = true;
+    }
   
   // Color specification 
-  // (cause XRender errors under many versions of Qt/X11)
-#ifndef Q_WS_X11
+  // (cause XRender errors under many versions of Qt4/X11)
+#ifndef Q_WS_X11 
   QApplication::setColorSpec(QApplication::ManyColor);
 #endif
   
   // Plugin mode
-#ifdef Q_WS_X11
+#if WITH_DJVIEWPLUGIN
   if (argc==2 && !strcmp(argv[1],"-netscape"))
     {
       verbose = true;
@@ -362,7 +436,6 @@ main(int argc, char *argv[])
 #endif
   
   // Discard session
-#ifdef Q_WS_X11
   if (argc==3 && !strcmp(argv[1],"-discard"))
     {
       QSettings s;
@@ -370,13 +443,11 @@ main(int argc, char *argv[])
       s.sync();
       return 0;
     }
-#endif
   
   // Create application
   QDjViewApplication app(argc, argv);
 
   // Restore session
-#ifdef Q_WS_X11
   if (app.isSessionRestored())
     {
       QSettings s;
@@ -396,13 +467,14 @@ main(int argc, char *argv[])
           return app.exec();
         }
     }
-#endif
   
   // Process command line
+  QStringList args;
   QDjView *main = app.newWindow();
   while (argc > 1 && argv[1][0] == '-')
     {
       QString arg = QString::fromLocal8Bit(argv[1]).replace(QRegExp("^-+"),"");
+      QString key = arg.section(QChar('='),0,1);
       if (arg == "help")
         usage();
       else if (arg == "verbose")
@@ -412,7 +484,7 @@ main(int argc, char *argv[])
       else if (arg == "fix")
         message(QApplication::tr("Option '-fix' is deprecated."));
       else 
-        message(main->parseArgument(arg));
+        args += arg;
       argc --;
       argv ++;
     }
@@ -435,7 +507,9 @@ main(int argc, char *argv[])
         }
     }
   
-  // Process events
+  // Process options and go
+  foreach(QString arg, args)
+    message(main->parseArgument(arg));
   main->show();
   return app.exec();
 }
