@@ -19,8 +19,9 @@
 # include "config.h"
 #endif
 
-#include "stdlib.h"
-#include "math.h"
+#include <stddef.h>
+#include <stdlib.h>
+#include <math.h>
 
 #include <QAbstractItemDelegate>
 #include <QAbstractListModel>
@@ -57,6 +58,9 @@
 #include <QTreeWidget>
 #include <QVariant>
 #include <QVBoxLayout>
+#if QT_VERSION >= 0x50000
+# include <QUrlQuery>
+#endif
 
 #include <libdjvu/ddjvuapi.h>
 #include <libdjvu/miniexp.h>
@@ -147,13 +151,11 @@ QDjViewOutline::refresh()
           tree->clear();
           QTreeWidgetItem *root = new QTreeWidgetItem();
           fillItems(root, miniexp_cdr(outline));
-          QTreeWidgetItem *item = root;
-          while (item->childCount()==1 &&
-                 item->data(0,Qt::UserRole).toInt() >= 0)
-            item = item->child(0);
-          while (item->childCount() > 0)
+          while (root->childCount() > 0)
             tree->insertTopLevelItem(tree->topLevelItemCount(),
-                                     item->takeChild(0) );
+                                     root->takeChild(0) );
+          if (tree->topLevelItemCount() == 1)
+            tree->topLevelItem(0)->setExpanded(true);
           delete root;
         }
       else
@@ -169,8 +171,9 @@ QDjViewOutline::refresh()
               QString name = djview->pageName(pageno);
               item->setText(0, tr("Page %1").arg(name));
               item->setData(0, Qt::UserRole, pageno);
+              item->setData(0, Qt::UserRole+1, pageno);
               item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
-              item->setToolTip(0, tr("Go to page %1").arg(name));
+              item->setToolTip(0, tr("Go: page %1.").arg(name));
               item->setWhatsThis(0, whatsThis());
             }
           tree->setItemExpanded(root, true);
@@ -180,6 +183,27 @@ QDjViewOutline::refresh()
 }
 
 
+int 
+QDjViewOutline::pageNumber(const char *link)
+{
+  if (link && link[0] == '#')
+    return djview->pageNumber(QString::fromUtf8(link+1));
+  if (link == 0 || link[0] != '?')
+    return -1;
+  QByteArray burl = QByteArray("http://f/f") + link;
+#if QT_VERSION >= 0x50000
+  QUrlQuery qurl(QUrl::fromEncoded(burl));
+#else
+  QUrl qurl = QUrl::fromEncoded(burl);
+#endif
+  if (qurl.hasQueryItem("page"))
+    return djview->pageNumber(qurl.queryItemValue("page"));
+  else if (qurl.hasQueryItem("pageno"))
+    return djview->pageNumber("$" + qurl.queryItemValue("pageno"));
+  return -1;
+}
+
+
 void 
 QDjViewOutline::fillItems(QTreeWidgetItem *root, miniexp_t expr)
 {
@@ -195,28 +219,25 @@ QDjViewOutline::fillItems(QTreeWidgetItem *root, miniexp_t expr)
           // fill item
           const char *name = miniexp_to_str(miniexp_car(s));
           const char *link = miniexp_to_str(miniexp_cadr(s));
-          int pageno = -1;
-          if (link && link[0]=='#')
-            pageno = djview->pageNumber(QString::fromUtf8(link+1));
+          int pageno = pageNumber(link);
+          QString pagename = (pageno>=0)?djview->pageName(pageno):QString();
           QTreeWidgetItem *item = new QTreeWidgetItem(root);
           if (name && name[0])
             item->setText(0, QString::fromUtf8(name));
-          else if (pageno >= 0)
-            item->setText(0, tr("Page %1").arg(djview->pageName(pageno)));
+          else if (! pagename.isEmpty())
+            item->setText(0, tr("Page %1").arg(pagename));
           item->setFlags(0);
           item->setWhatsThis(0, whatsThis());
-          if (link && link[0]=='#' && pageno>=0)
+          if (link && link[0])
             {
-              item->setData(0, Qt::UserRole, pageno);
+              QString slink = QString::fromUtf8(link);
+              item->setData(0, Qt::UserRole+1, slink);
               item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
-              QString pagename = djview->pageName(pageno);
-              item->setToolTip(0, tr("Go to page %1").arg(pagename));
-            } 
-          else if (link && link[0])
-            {
-              item->setData(0, Qt::UserRole, QString::fromUtf8(link));
-              item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
-              item->setToolTip(0, QString::fromUtf8(link));
+              item->setToolTip(0, tr("Go: %1").arg(slink));
+              if (pageno >= 0)
+                item->setData(0, Qt::UserRole, pageno);
+              if (! pagename.isEmpty())
+                item->setToolTip(0, tr("Go: page %1.").arg(pagename));
             }
           // recurse
           fillItems(item, miniexp_cddr(s));
@@ -275,19 +296,19 @@ QDjViewOutline::searchItem(QTreeWidgetItem *item, int pageno,
 void 
 QDjViewOutline::itemActivated(QTreeWidgetItem *item)
 {
-  QVariant data = item->data(0, Qt::UserRole);
-  if (data.type() == QVariant::Int)
-    {
-      int page = data.toInt();
-      if (page >= 0 && page < djview->pageNum())
-        djview->goToPage(page);
-    }
-  else if (data.type() == QVariant::String)
+  QVariant data = item->data(0, Qt::UserRole+1);
+  if (data.type() == QVariant::String)
     {
       QString link = data.toString();
       if (link.size() > 0)
         djview->goToLink(link);
     }
+  else if (data.type() == QVariant::Int)
+    {
+      int pageno = data.toInt();
+      if (pageno >= 0)
+        djview->goToPage(pageno);
+    }
 }
 
 
@@ -312,7 +333,6 @@ protected:
   QStyleOptionViewItem viewOptions() const;
 private:
   QDjViewThumbnails *widget;
-  QDjView *djview;
 };
 
 
@@ -357,8 +377,7 @@ private:
 
 QDjViewThumbnails::View::View(QDjViewThumbnails *widget)
   : QListView(widget), 
-    widget(widget), 
-    djview(widget->djview)
+    widget(widget)
 {
   setDragEnabled(false);
   setEditTriggers(QAbstractItemView::NoEditTriggers);
@@ -713,7 +732,7 @@ QDjViewThumbnails::QDjViewThumbnails(QDjView *djview)
   action->setData(true);
   updateActions();
 
-#ifdef Q_WS_MAC
+#ifdef Q_OS_DARWIN
   QString mc = tr("Control Left Mouse Button");
 #else
   QString mc = tr("Right Mouse Button");
@@ -748,7 +767,7 @@ QDjViewThumbnails::updateActions(void)
 void 
 QDjViewThumbnails::pageChanged(int pageno)
 {
-  if (pageno>=0 && pageno<djview->pageNum())
+  if (pageno >= 0 && pageno < djview->pageNum())
     {
       QModelIndex mi = model->index(pageno);
       if (! selection->isSelected(mi))
@@ -1126,6 +1145,21 @@ parse_text_type(miniexp_t exp)
   return -1;
 }
 
+static int count_final_spaces(const QString &s)
+{
+  int c = 0;
+  int l = s.size();
+  while (l>0 && s[--l].isSpace())
+    c += 1;
+  return c;
+}
+
+static QString chopped(QString s, int c)
+{
+  s.chop(c);
+  return s;
+}
+
 static bool
 miniexp_get_text(miniexp_t exp, QString &result, 
                  QMap<int,miniexp_t> &positions, int &state)
@@ -1142,10 +1176,14 @@ miniexp_get_text(miniexp_t exp, QString &result,
   state = qMax(state, typenum);
   if (miniexp_stringp(s) && !miniexp_cdr(r))
     {
-      result += (state >= 2) ? "\n" : (state >= 1) ? " " : "";
+      int c = count_final_spaces(result);
+      if (state >= 2 && !result.endsWith('\n'))
+        result = chopped(result,c) + "\n";
+      else if (state >= 1 && c == 0)
+        result += " ";
       state = -1;
       positions[result.size()] = exp;
-      result += QString::fromUtf8(miniexp_to_str(s));
+      result += QString::fromUtf8(miniexp_to_str(s)).trimmed();
       r = miniexp_cdr(r);
     }
   while(miniexp_consp(s))
@@ -1392,7 +1430,7 @@ QDjViewFind::Model::animTimeout()
 {
   if (animButton && !animIcon.isNull())
     {
-      if (animButton->icon().serialNumber() == findIcon.serialNumber())
+      if (animButton->icon().cacheKey() == findIcon.cacheKey())
         animButton->setIcon(animIcon);
       else
         animButton->setIcon(findIcon);
@@ -1403,6 +1441,7 @@ QDjViewFind::Model::animTimeout()
 void 
 QDjViewFind::Model::nextHit(bool backwards)
 {
+  djview->getDjVuWidget()->terminateAnimation();
   if (working && backwards != searchBackwards)
     {
       pending = false;
@@ -1472,6 +1511,9 @@ QDjViewFind::Model::makeSelectionVisible()
 void 
 QDjViewFind::Model::pageChanged(int pageno)
 {
+  QDjVuWidget *djvu = djview->getDjVuWidget();
+  if (djvu && djvu->animationInProgress())
+    return;
   if (pageno != curPage)
     {
       curHit = -1;
@@ -1519,8 +1561,13 @@ QDjViewFind::Model::itemActivated(const QModelIndex &mi)
   if (mi.isValid())
     {
       int row = mi.row();
-      if (row>=0 && row<pages.size())
-        djview->goToPage(pages[row].pageno);
+      if (row >= 0 && row < pages.size() && pages[row].hits > 0)
+        {
+          curPage = pages[row].pageno;
+          curHit = -1;
+          pending = true;
+          doPending();
+        }
     }
 }
 
@@ -1798,7 +1845,7 @@ void
 QDjViewFind::findNext()
 {
   if (text().isEmpty())
-    djview->showSideBar("find");
+    djview->showFind();
   model->nextHit(false);
 }
 
@@ -1807,7 +1854,7 @@ void
 QDjViewFind::findPrev()
 {
   if (text().isEmpty())
-    djview->showSideBar("find");
+    djview->showFind();
   model->nextHit(true);
 }
 
@@ -1825,7 +1872,7 @@ QDjViewFind::findAgain()
 void 
 QDjViewFind::pageChanged(int pageno)
 {
-  if (pageno>=0 && pageno<djview->pageNum())
+  if (pageno >= 0 && pageno < djview->pageNum())
     {
       model->pageChanged(pageno);
       if (model->modelSelect(pageno))
